@@ -7,7 +7,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -87,7 +87,7 @@ check_bits (gnutls_session_t session, gnutls_x509_crt_t crt, unsigned int max_bi
 
 /* three days */
 #define MAX_OCSP_VALIDITY_SECS (3*60*60*24)
-
+#ifdef ENABLE_OCSP
 /* If the certificate is revoked status will be GNUTLS_CERT_REVOKED.
  * 
  * Returns:
@@ -198,6 +198,7 @@ cleanup:
   return ret;
 }
 
+#endif
 
 #define CLEAR_CERTS for(x=0;x<peer_certificate_list_size;x++) { \
 	if (peer_certificate_list[x]) \
@@ -321,12 +322,14 @@ _gnutls_x509_cert_verify_peers (gnutls_session_t session,
         }
     }
 
+#ifdef ENABLE_OCSP
   ret = check_ocsp_response(session, peer_certificate_list[0], issuer, &resp, &ocsp_status);
   if (ret < 0)
     {
       CLEAR_CERTS;
       return gnutls_assert_val(ret);
     }
+#endif
 
 skip_ocsp:
   /* Verify certificate 
@@ -623,13 +626,28 @@ read_cert_mem (gnutls_certificate_credentials_t res, const void *cert,
   return ret;
 }
 
+static int tmp_pin_cb(void* userdata, int attempt, const char *token_url, const char *token_label,
+	unsigned int flags, char *pin, size_t pin_max)
+{
+const char* tmp_pin = userdata;
+
+  if (attempt == 0) 
+    {
+      snprintf(pin, pin_max, "%s", tmp_pin);
+      return 0;
+    }
+  
+  return -1;
+}
+
 /* Reads a PEM encoded PKCS-1 RSA/DSA private key from memory.  Type
  * indicates the certificate format.  KEY can be NULL, to indicate
  * that GnuTLS doesn't know the private key.
  */
 static int
 read_key_mem (gnutls_certificate_credentials_t res,
-              const void *key, int key_size, gnutls_x509_crt_fmt_t type)
+              const void *key, int key_size, gnutls_x509_crt_fmt_t type,
+              const char* pass, unsigned int flags)
 {
   int ret;
   gnutls_datum_t tmp;
@@ -649,8 +667,13 @@ read_key_mem (gnutls_certificate_credentials_t res,
       
       if (res->pin.cb)
         gnutls_privkey_set_pin_function(privkey, res->pin.cb, res->pin.data);
+      else if (pass != NULL)
+        {
+          snprintf(res->pin_tmp, sizeof(res->pin_tmp), "%s", pass);
+          gnutls_privkey_set_pin_function(privkey, tmp_pin_cb, res->pin_tmp);
+        }
 
-      ret = gnutls_privkey_import_x509_raw (privkey, &tmp, type, NULL, 0);
+      ret = gnutls_privkey_import_x509_raw (privkey, &tmp, type, pass, flags);
       if (ret < 0)
         {
           gnutls_assert ();
@@ -721,8 +744,6 @@ cleanup:
 }
 
 #ifdef ENABLE_PKCS11
-/* Reads a private key from a token.
- */
 static int
 read_cas_url (gnutls_certificate_credentials_t res, const char *url)
 {
@@ -734,7 +755,7 @@ read_cas_url (gnutls_certificate_credentials_t res, const char *url)
   /* FIXME: should we use login? */
   ret =
     gnutls_pkcs11_obj_list_import_url (NULL, &pcrt_list_size, url,
-                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED, 0);
+                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED_CA, 0);
   if (ret < 0 && ret != GNUTLS_E_SHORT_MEMORY_BUFFER)
     {
       gnutls_assert ();
@@ -756,7 +777,7 @@ read_cas_url (gnutls_certificate_credentials_t res, const char *url)
 
   ret =
     gnutls_pkcs11_obj_list_import_url (pcrt_list, &pcrt_list_size, url,
-                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED, 0);
+                                       GNUTLS_PKCS11_OBJ_ATTR_CRT_TRUSTED_CA, 0);
   if (ret < 0)
     {
       gnutls_assert ();
@@ -869,6 +890,9 @@ cleanup:
   gnutls_free (ccert);
   return ret;
 }
+#else
+# define read_cert_url(x,y) gnutls_assert_val(GNUTLS_E_UNIMPLEMENTED_FEATURE)
+# define read_cas_url(x,y) gnutls_assert_val(GNUTLS_E_UNIMPLEMENTED_FEATURE)
 #endif
 
 /* Reads a certificate file
@@ -881,12 +905,10 @@ read_cert_file (gnutls_certificate_credentials_t res,
   size_t size;
   char *data;
 
-#ifdef ENABLE_PKCS11
   if (strncmp (certfile, "pkcs11:", 7) == 0)
     {
       return read_cert_url (res, certfile);
     }
-#endif /* ENABLE_PKCS11 */
 
   data = read_binary_file (certfile, &size);
 
@@ -910,16 +932,20 @@ read_cert_file (gnutls_certificate_credentials_t res,
  */
 static int
 read_key_file (gnutls_certificate_credentials_t res,
-               const char *keyfile, gnutls_x509_crt_fmt_t type)
+               const char *keyfile, gnutls_x509_crt_fmt_t type, const char* pass,
+               unsigned int flags)
 {
   int ret;
   size_t size;
   char *data;
 
-  if (gnutls_url_is_supported(keyfile))
+  if (_gnutls_url_is_known(keyfile)) 
     {
-      return read_key_url (res, keyfile);
-    }
+      if (gnutls_url_is_supported(keyfile))
+        return read_key_url (res, keyfile);
+      else
+        return gnutls_assert_val(GNUTLS_E_UNIMPLEMENTED_FEATURE);
+  }
 
   data = read_binary_file (keyfile, &size);
 
@@ -929,7 +955,7 @@ read_key_file (gnutls_certificate_credentials_t res,
       return GNUTLS_E_FILE_ERROR;
     }
 
-  ret = read_key_mem (res, data, size, type);
+  ret = read_key_mem (res, data, size, type, pass, flags);
   free (data);
 
   return ret;
@@ -964,13 +990,50 @@ gnutls_certificate_set_x509_key_mem (gnutls_certificate_credentials_t res,
                                      const gnutls_datum_t * cert,
                                      const gnutls_datum_t * key,
                                      gnutls_x509_crt_fmt_t type)
+{
+  return gnutls_certificate_set_x509_key_mem2(res, cert, key, type, NULL, 0);
+}
+
+/**
+ * gnutls_certificate_set_x509_key_mem2:
+ * @res: is a #gnutls_certificate_credentials_t structure.
+ * @cert: contains a certificate list (path) for the specified private key
+ * @key: is the private key, or %NULL
+ * @type: is PEM or DER
+ * @pass: is the key's password
+ * @flags: an ORed sequence of gnutls_pkcs_encrypt_flags_t
+ *
+ * This function sets a certificate/private key pair in the
+ * gnutls_certificate_credentials_t structure. This function may be called
+ * more than once, in case multiple keys/certificates exist for the
+ * server.
+ *
+ * Note that the keyUsage (2.5.29.15) PKIX extension in X.509 certificates
+ * is supported. This means that certificates intended for signing cannot
+ * be used for ciphersuites that require encryption.
+ *
+ * If the certificate and the private key are given in PEM encoding
+ * then the strings that hold their values must be null terminated.
+ *
+ * The @key may be %NULL if you are using a sign callback, see
+ * gnutls_sign_callback_set().
+ *
+ * Returns: %GNUTLS_E_SUCCESS (0) on success, or a negative error code.
+ **/
+int
+gnutls_certificate_set_x509_key_mem2 (gnutls_certificate_credentials_t res,
+                                     const gnutls_datum_t * cert,
+                                     const gnutls_datum_t * key,
+                                     gnutls_x509_crt_fmt_t type,
+                                     const char* pass,
+                                     unsigned int flags)
 {
   int ret;
 
   /* this should be first
    */
   if ((ret = read_key_mem (res, key ? key->data : NULL,
-                           key ? key->size : 0, type)) < 0)
+                           key ? key->size : 0, type, pass, flags)) < 0)
     return ret;
 
   if ((ret = read_cert_mem (res, cert->data, cert->size, type)) < 0)
@@ -1300,12 +1363,46 @@ gnutls_certificate_set_x509_key_file (gnutls_certificate_credentials_t res,
                                       const char *certfile,
                                       const char *keyfile,
                                       gnutls_x509_crt_fmt_t type)
+{
+  return gnutls_certificate_set_x509_key_file2(res, certfile, keyfile, type, NULL, 0);
+}
+
+/**
+ * gnutls_certificate_set_x509_key_file2:
+ * @res: is a #gnutls_certificate_credentials_t structure.
+ * @certfile: is a file that containing the certificate list (path) for
+ *   the specified private key, in PKCS7 format, or a list of certificates
+ * @keyfile: is a file that contains the private key
+ * @type: is PEM or DER
+ * @pass: is the password of the key
+ * @flags: an ORed sequence of gnutls_pkcs_encrypt_flags_t
+ *
+ * This function sets a certificate/private key pair in the
+ * gnutls_certificate_credentials_t structure.  This function may be
+ * called more than once, in case multiple keys/certificates exist for
+ * the server.  For clients that need to send more than its own end
+ * entity certificate, e.g., also an intermediate CA cert, then the
+ * @certfile must contain the ordered certificate chain.
+ *
+ * This function can also accept URLs at @keyfile and @certfile. In that case it
+ * will import the private key and certificate indicated by the URLs. Note
+ * that the supported URLs are the ones indicated by gnutls_url_is_supported().
+ *
+ * Returns: %GNUTLS_E_SUCCESS (0) on success, or a negative error code.
+ **/
+int
+gnutls_certificate_set_x509_key_file2 (gnutls_certificate_credentials_t res,
+                                      const char *certfile,
+                                      const char *keyfile,
+                                      gnutls_x509_crt_fmt_t type,
+                                      const char* pass,
+                                      unsigned int flags)
 {
   int ret;
 
   /* this should be first
    */
-  if ((ret = read_key_file (res, keyfile, type)) < 0)
+  if ((ret = read_key_file (res, keyfile, type, pass, flags)) < 0)
     return ret;
 
   if ((ret = read_cert_file (res, certfile, type)) < 0)
@@ -1361,7 +1458,7 @@ add_new_crt_to_rdn_seq (gnutls_certificate_credentials_t res, gnutls_x509_crt_t*
           return GNUTLS_E_SHORT_MEMORY_BUFFER;
         }
 
-      newdata = gnutls_realloc (res->x509_rdn_sequence.data, newsize);
+      newdata = gnutls_realloc_fast (res->x509_rdn_sequence.data, newsize);
       if (newdata == NULL)
         {
           gnutls_assert ();
@@ -1665,12 +1762,10 @@ gnutls_certificate_set_x509_trust_file (gnutls_certificate_credentials_t cred,
   gnutls_datum_t cas;
   size_t size;
 
-#ifdef ENABLE_PKCS11
   if (strncmp (cafile, "pkcs11:", 7) == 0)
     {
       return read_cas_url (cred, cafile);
     }
-#endif
 
   cas.data = (void*)read_binary_file (cafile, &size);
   if (cas.data == NULL)
@@ -2150,3 +2245,41 @@ void gnutls_certificate_set_pin_function (gnutls_certificate_credentials_t cred,
   cred->pin.cb = fn;
   cred->pin.data = userdata;
 }
+
+/**
+ * gnutls_url_is_supported:
+ * @url: A PKCS 11 url
+ *
+ * Check whether url is supported.  Depending on the system libraries
+ * GnuTLS may support pkcs11 or tpmkey URLs.
+ *
+ * Returns: return non-zero if the given URL is supported, and zero if
+ * it is not known.
+ *
+ * Since: 3.1.0
+ **/
+int
+gnutls_url_is_supported (const char* url)
+{
+#ifdef ENABLE_PKCS11
+  if (strstr(url, "pkcs11:") != NULL)
+    return 1;
+#endif
+#ifdef HAVE_TROUSERS
+  if (strstr(url, "tpmkey:") != NULL)
+    return 1;
+#endif
+  return 0;
+}
+
+int
+_gnutls_url_is_known (const char* url)
+{
+  if (strstr(url, "pkcs11:") != NULL)
+    return 1;
+  else if (strstr(url, "tpmkey:") != NULL)
+    return 1;
+  else
+    return 0;
+}
+
